/*
    pool.h - linear tree and collision pool handling (kd-tree-2)
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __KDTREE_POOL_H
#define __KDTREE_POOL_H
#include <stdint.h>
#include "space.h"
#include "kd-tree.h"

/*
  The kd-tree element pool functions handle the data structures
  (linear kd-tree, element pool, pair pool) holding all elements
  and sub-elements information.
  
  These functions replace those in "array.h".
*/

// *********************************************************************
// Type options
// *********************************************************************
/*
  The following integer types are used to count structures of different types
  (KDTREE_INDEX: kdtree_element, KDTREE_PAIR: kdtree_pair, KDTREE_STATS: /,
  KDTREE_ENTRIES: kdtree_entry). KDTREE_ENTRIES may be made signed for enhanced
  sort speed, at the cost of one bit.
*/
#define KDTREE_INDEX	uint32_t
#define KDTREE_PAIRS	uint32_t
#define KDTREE_STATS	int32_t

// #define KDTREE_ENTRIES_IS_UNSIGNED
#define KDTREE_ENTRIES	int32_t

/*
  Both kdtree_entry and kdtree_pair structures hold references to
  kdtree_element structures. These references can be either a KDTREE_INDEX
  indicating an offset in the kdtree_pool, or a direct pointer.
  Both choices have their advantages and disadvantages, but exhibit comparable
  performance in terms of CPU time. Therefore, the main issue is memory usage,
  depending on what are the sizes of pointers and KDTREE_INDEXes on your
  particular host machine.
  
  Constraint: if pairs are indexed, then entries must be indexed too. All
  three other combinations are valid.
*/
#define KDTREE_INDEXED_ENTRIES
#define KDTREE_INDEXED_PAIRS

// *********************************************************************
// Data structures
// *********************************************************************
/*
  A kdtree_element uniquely describes an element of the simulations. There
  should be one and only one instance of kdtree_element for each element:
  pointer comparisons are used to identify elements.
  
  A kdtree_pair is created when two bounding volumes intersect each other. It
  holds references to the respective kdtree_elements. Its purpose is to
  have each such pair of elements appear only once in an array. In addition,
  each kdtree_pair belongs to two linked lists. These lists contain all
  pairs implying a kdtree_element.
  
  A kdtree_pair_buffer describes an array of kdtree_pair. Pair buffers are
  organised in a linked list for the (likely) case one is not enough.
  
  A kdtree_entry is an occurence (among possibly several others) of a
  kdtree_element in the linear kd-tree. It holds a reference to an element
  and the locational code of a node in the tree.
  
  A kdtree_pool gathers all broad-phase data, i.e. an array of kdtree_element,
  an array of kdtree_entry (the linear kd-tree itself), and a list of
  kdtree_pair_buffer.
  
*/
typedef struct kdtree_element_t kdtree_element;
typedef struct kdtree_pair_t kdtree_pair;
typedef struct kdtree_pair_buffer_t kdtree_pair_buffer;
typedef struct kdtree_entry_t kdtree_entry;
typedef struct kdtree_pool_t kdtree_pool;

struct kdtree_element_t {
	void *pointer;
	kdtree_sphere sphere;
	kdtree_pair *first;
};

#ifdef KDTREE_INDEXED_PAIRS
struct kdtree_pair_t {
	KDTREE_INDEX idx_A, idx_B;
	kdtree_pair *next_A, *next_B;
};
#else
struct kdtree_pair_t {
	kdtree_element *element_A, *element_B;
	kdtree_pair *next_A, *next_B;
};
#endif

struct kdtree_pair_buffer_t {
	KDTREE_PAIRS npairs;
	kdtree_pair_buffer *next;
	kdtree_pair pairs[];
};

#ifdef KDTREE_INDEXED_ENTRIES
struct kdtree_entry_t {
	kdtree_node path;
	KDTREE_INDEX idx;
};
#else
struct kdtree_entry_t {
	kdtree_node path;
	kdtree_element *element;
};
#endif

struct kdtree_pool_t {
	KDTREE_INDEX nel, nel_alloc;
	KDTREE_ENTRIES nent;

	kdtree_element *elements;
	kdtree_entry *entries;
	
	KDTREE_PAIRS pair_gran;
	kdtree_pair_buffer *pair_buf;
	
	KDTREE_STATS stats_out_of_bounds;
	KDTREE_STATS stats_tree_collisions;
	KDTREE_STATS stats_sphere_collisions;
	KDTREE_STATS stats_unique_collisions;
};


// *********************************************************************
// Functions
// *********************************************************************
kdtree_pool *kdtree_pool_new(KDTREE_INDEX nelements, KDTREE_PAIRS pairs_base);
void kdtree_pool_free(kdtree_pool *b);


// Offline linear kd-tree
int  kdtree_pool_add_sphere(kdtree_pool *b, const kdtree_sphere *bs, void *obj);
void kdtree_pool_build_tree(kdtree_pool *b, const kdtree_space *s);
void kdtree_pool_sort_tree(kdtree_pool *b);
void kdtree_pool_sweep_tree(kdtree_pool *b);


// *********************************************************************
// Helper defines
// *********************************************************************
#ifdef KDTREE_INDEXED_PAIRS

#define KDTREE_PAIR_ELEMENT_A(b, buf, i) \
	((b)->elements[(buf)->pairs[i].idx_A])
#define KDTREE_PAIR_ELEMENT_B(b, buf, i) \
	((b)->elements[(buf)->pairs[i].idx_B])

#else

#define KDTREE_PAIR_ELEMENT_A(b, buf, i) \
	(*((buf)->pairs[i].element_A))
#define KDTREE_PAIR_ELEMENT_B(b, buf, i) \
	(*((buf)->pairs[i].element_B))

#endif




#endif

