/*
    sort_pool.c - linear tree quicksort for pool.c
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include "kd-tree.h"
#include "pool.h"
#include "sort_pool.h"

#include <stdio.h>
#define DEBUG(s...)	fprintf(stderr, s)

// ********************************************************************
//
// ********************************************************************
static void kdtree_ent_swap(kdtree_entry *const tree,
	const KDTREE_ENTRIES i, const KDTREE_ENTRIES j)
{
	kdtree_entry tmp;
	
	if (i != j) {
		memcpy(&tmp, &tree[i], sizeof(tmp));
		memcpy(&tree[i], &tree[j], sizeof(tmp));
		memcpy(&tree[j], &tmp, sizeof(tmp));
	}
}

// ********************************************************************
//
// ********************************************************************
static KDTREE_ENTRIES kdtree_qs_partition(kdtree_entry *const tree,
	const KDTREE_ENTRIES left, const KDTREE_ENTRIES right,
	const KDTREE_ENTRIES pivot)
{
	KDTREE_ENTRIES r, i;
	
	// place pivot at right end, and store its value
	kdtree_ent_swap(tree, pivot, right);
	
	r = left;
	
	for (i = left; i < right; i++) {
		// accumulate values least than pivot_value at left end
		// (before index r)
		if (KDTREE_BEFORE(tree[i].path, tree[right].path)) {
			kdtree_ent_swap(tree, i, r);
			r++;
		}
	}
	kdtree_ent_swap(tree, r, right);
	return(r);
}

// ********************************************************************
//
// ********************************************************************
// INSERTION_THRESHOLD must be >= 1 or we might break quicksort
#define INSERTION_THRESHOLD	32
static void kdtree_pool_insertion_sort(kdtree_entry *const tree,
	const KDTREE_ENTRIES left, const KDTREE_ENTRIES right)
{
	kdtree_entry tmp;
	KDTREE_ENTRIES k, i;
	
	for (k = left + 1; k <= right; k++) {
		memcpy(&tmp, &tree[k], sizeof(tmp));
		
		i = k - 1;
		while ((i >= left)
#ifdef KDTREE_ENTRIES_IS_UNSIGNED
		&& (i != (KDTREE_ENTRIES)-1)
#endif
		&& (KDTREE_BEFORE(tmp.path, tree[i].path))) {
			memcpy(&tree[i + 1], &tree[i], sizeof(tmp));
			i--;
		}
		
		memcpy(&tree[i + 1], &tmp, sizeof(tmp));
	}
}


// ********************************************************************
//
// ********************************************************************
void kdtree_pool_quicksort(kdtree_entry *tree,
	KDTREE_ENTRIES left, KDTREE_ENTRIES right)
{
	KDTREE_ENTRIES pivot1;
	
	while (left + INSERTION_THRESHOLD <= right) {
		pivot1 = kdtree_qs_partition(tree, left, right,
			(left + right) >> 1);

		if (left + 1 < pivot1)
			kdtree_pool_quicksort(tree, left, pivot1 - 1);
		
		// A terminal loop replaces the second recursion
		left = pivot1 + 1;
	}
	
	kdtree_pool_insertion_sort(tree, left, right);
}

